
------------------------------------------------------------------------
--
-- ShieldFrames
--
-- Copyright (C) 2010  Philipp Schmidt
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to:
--
-- Free Software Foundation, Inc.,
-- 51 Franklin Street, Fifth Floor,
-- Boston, MA  02110-1301, USA.
--
--
------------------------------------------------------------------------


local ShieldFrames = _G.ShieldFrames;
if(not ShieldFrames) then return; end

local AbsorbsMonitor = LibStub("LibAbsorbsMonitor-1.0");

local AnchorProto = {};
ShieldFrames.AnchorProto = AnchorProto;

local AceRegistry = LibStub("AceConfigRegistry-3.0");

local playerGUID;
local UnitGUID = _G.UnitGUID;

-- Shared across anchors
local AreaTriggerSpells = {};


local backdrop = {bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
		edgeFile = "Interface\\ChatFrame\\ChatFrameBackground", edgeSize = 0.90,
		insets = {left = 1, right = 1, top = 1, bottom = 1}};
		
		
-- #################################################################################################


local function GetUnitClassColorRGB(name)
	local _, class = UnitClass(name);

	if _G.RAID_CLASS_COLORS and _G.RAID_CLASS_COLORS[class] then
		return _G.RAID_CLASS_COLORS[class].r, _G.RAID_CLASS_COLORS[class].g, _G.RAID_CLASS_COLORS[class].b;
	else
		return 0.63, 0.63, 0.63;	
	end
end

local function NoOp()
end

local function OnDragStop(self)
	self:StopMovingOrSizing();

	local x, y = self:GetLeft(), self:GetTop();
		
	self.profile.posX, self.profile.posY = x, y;
	
	AceRegistry:NotifyChange("ShieldFrames");
end


-- #################################################################################################


local function unitRelativeFilter_CheckInclusive(self, guid)
	for k, v in pairs(self.profile.unitRelativeFilter) do
		if(UnitGUID(k) == guid) then
			return true;
		end
	end
	
	return false;
end

local function unitRelativeFilter_CheckExclusive(self, guid)
	for k, v in pairs(self.profile.unitRelativeFilter) do
		if(UnitGUID(k) == guid) then
			return false;
		end
	end
	
	return true;
end


-- #################################################################################################


local function ungrouped_ReanchorFrames(self)
	local reverse = self.profile.reverseGrowth;
	local cur = self.next;

	-- circular reference protection
	local i = 0;
	
	while(cur) do
		if(i == 500) then
			error("Loop counter on reanchoring reached 500, circular effect reference?");
		end
		
		i = i + 1;
		
		cur:AlignTo(cur.prev, reverse);
		cur = cur.next;
	end
end

local function ungrouped_NewEffect(self, effect, guid, name, spellId)
	effect.prev = self.lastChild;
	self.lastChild.next = effect;
	effect.next = nil;
		
	effect:AlignTo(self.lastChild, self.profile.reverseGrowth);
	
	self.lastChild = effect;
end

local function ungrouped_RemoveEffect(self, effect, guid, spellId)
	local prev = effect.prev;
	local next = effect.next;
		
	if(self.lastChild == effect) then
		self.lastChild = prev;
	end
		
	prev.next = next;
		
	if(next) then
		next.prev = prev;		
		next:AlignTo(prev, self.profile.reverseGrowth);
	end
end


-- #################################################################################################


-- Linkage on the unit is down via:
--
-- prevHeader, nextHeader
-- next (the first effect, like for the anchor)
-- lastChild

local function byUnit_ApplyProfile(self)
	for guid, unit in pairs(self.units) do
		unit:ApplyProfile();
	end
end
		
local function byUnit_ReanchorAll(self)	
	local prevHeaderAnchor = self;	
	local curEffect;	
	
	-- circular reference protection
	local i = 0;	
	
	-- First we need to get to the first unit. While we could have saved this
	-- on creation time in self.nextHeader, we only need it here.
	-- So we simply traverse the list backwards.
	
	local curUnit = self.lastChild;
	
	while(curUnit.prevHeader ~= self) do
		if(i == 500) then
			error("Loop counter on finding first unit reached 500, circular unit reference?");
		end	
		
		i = i + 1;
		
		curUnit = curUnit.prevHeader;
	end
	
	i = 0;
	
	while(curUnit) do		
		curUnit:AlignTo(prevHeaderAnchor);			
		
		curEffect = curUnit.next;		
		
		while(curEffect) do
			if(i == 500) then
				error("Loop counter on reanchoring reached 500, circular unit/effect reference?");
			end	
		
			i = i + 1;
			
			curEffect:AlignTo(curEffect.prev);
					
			prevHeaderAnchor = curEffect;			
			curEffect = curEffect.next;
		end
		
		curUnit = curUnit.nextHeader;
	end
end

local function byUnit_ReanchorEffects(self, guid, spellId)
	if(not guid) then
		return byUnit_ReanchorAll(self);
	end
	
	local unit = self.units[guid];
	
	if(not unit) then return; end
	
	local cur = unit.next;
	
	-- circular reference protection
	local i = 0;
	
	while(cur) do
		if(i == 500) then
			error("Loop counter on reanchoring reached 500, circular effect reference?");
		end
		
		i = i + 1;
		
		cur:AlignTo(cur.prev);
		
		unit.lastChild = cur;
		cur = cur.next;		
	end
	
	if(unit.nextHeader) then
		unit.nextHeader:AlignTo(unit.lastChild);
		unit.lastChild.alignedHeader = unit.nextHeader;
	end
end

local function byUnit_SortEffects(self, guid, spellId)
	if(not guid) then
		for guid, guidEffects in pairs(self.effects) do
			byUnit_SortEffects(self, guid);
		end
		
		return;
	end
	
	local guidEffects = self.effects[guid];
	
	if(not guidEffects) then return; end
	
	local unit = self.units[guid];
	local effectsList = AbsorbsMonitor.Unit_EffectsList(guid);
		
	local prev = unit;	
	local cur;
	
	-- We got no list from AbsorbsMonitor
	-- This happens quite often if we're called
	-- on the trigger of an area effect.
	-- We put it into random order as given by the
	-- map. It's only important we have a working
	-- linkage after this function
	if(not effectsList) then		
		for k, v in pairs(guidEffects) do
			v.prev = prev;
			v.prev.next = v;
			v.next = nil;
			
			cur = v;
		end
		
		unit.lastChild = cur;
		
	else
		for i = #effectsList, 1, -1 do
			cur = guidEffects[effectsList[i][1]];
		
			cur.prev = prev;
			cur.next = nil;
			cur.alignedHeader = nil;
		
			prev.next = cur;		
		
			prev = cur;
		end	
			
		unit.lastChild = cur;
	end
	
	byUnit_ReanchorEffects(self, guid);
end

local function byUnit_NewEffect(self, effect, guid, name, spellId)	
	local unit = self.units[guid];

	if(not unit) then
		unit = ShieldFrames.HeaderProto:New(self);
		
		unit.leftText:SetText(name);
		unit.leftText:SetTextColor(GetUnitClassColorRGB(name));
		
		unit.rightText:SetText("");
		
		self.units[guid] = unit;
		
		self.lastChild.nextHeader = unit;
		unit.prevHeader = self.lastChild;		
		unit.nextHeader = nil;
		
		if(self.lastChild == self) then			
			unit:AlignTo(self);
			
		else
			local lastEffect = self.lastChild.lastChild;
			
			unit:AlignTo(lastEffect);			
			lastEffect.alignedHeader = unit;
		end
		
		self.lastChild = unit;
		
		unit.next = nil;
		unit.lastChild = nil;
	end
	
	-- Since we use sorting here, the reanchoring within the effect will happen there		
end
	
local function byUnit_RemoveEffect(self, effect, guid, spellId)
	local prev = effect.prev;
	local next = effect.next;
	local unit = self.units[guid];
	
	if(not unit) then return; end
	
	-- So this was the last effect of this unit
	-- We will remove the header with it
	if(unit.next == unit.lastChild) then
		--byUnit_RemoveHeader(self, guid);
				
		-- We have a following header
		if(unit.nextHeader) then
			-- We are aligned to the anchor
			if(unit.prevHeader == self) then
				unit.nextHeader:AlignTo(self);
				
			-- We are aligned at some other header's effect
			else
				unit.nextHeader:AlignTo(unit.prevHeader.lastChild);
				unit.prevHeader.lastChild.alignedHeader = unit.nextHeader;
				
				unit.prevHeader.nextHeader = unit.nextHeader;				
			end
			
			unit.nextHeader.prevHeader = unit.prevHeader;			
						
		-- No following header, this is the anchor's last child
		else 
			self.lastChild = unit.prevHeader;
			
			unit.prevHeader.lastChild.alignedHeader = nil;
			unit.prevHeader.nextHeader = nil;
		end
		
		unit.next = nil;
		unit.lastChild = unit;
		
		unit:Release();
		self.units[guid] = nil;
		
	-- This was not the last one
	else
		if(unit.next == effect) then
			unit.next = next	
		elseif(unit.lastChild == effect) then
			unit.lastChild = prev;
		end
		
		prev.next = next;
		
		if(next) then
			next.prev = prev;		
			next:AlignTo(prev);
			
		elseif(effect.alignedHeader) then			
			effect.alignedHeader:AlignTo(prev);
			
			prev.alignedHeader = effect.alignedHeader;
			effect.alignedHeader = nil;
		end
	end
end

local function byUnit_UnitUpdated(self, guid, value, quality)
	local unit = self.units[guid];
	
	if(not unit) then return; end
	
	unit.rightText:SetText(value);
	unit.rightText:SetTextColor(1 - quality, quality, 0);
end

local function byUnit_Clear(self)
	for guid, unit in pairs(self.units) do
		unit:Release();
	end
	
	self.units = {};
end


-- #################################################################################################

function AnchorProto.Enable()
	playerGUID = UnitGUID("player");
end

function AnchorProto.defaults()
	return {
		type = "ungrouped",
		posX = 600,
		posY = 600,
		scale = 1.0,
		width = 200,
		anchorHeight = 20,
		effectHeight = 16,
		anchorTextHeight = 10,
		durationHeight = 3,
		anchorAlpha = 1.0,
		iconVisible = true,
		durationVisible = true,
		effectAlpha = 1.0,
		effectSpacing = 3,
		effectTextSize = -3,
		leftFormat = "%d - %n",
		centerForamt = "",
		rightFormat = "%a (%p%)",
		anchorVisible = true,
		filterAppliedByMe = false,
		filterAppliedOnMe = false,
		unitFilterMode = "exclusive",
		unitRelativeFilter = {
		},
		unitSpecificFilter = {
		},
		spellFilterMode = "exclusive",
		spellFilter = {
		},
		headerVisible = true,
		headerAlpha = 1.0,
		headerHeight = 20,
		headerSpacing = 5,
		headerTextSize = -1,
		reverseGrowth = false,
	};
end

function AnchorProto:New(name, profile)
	local o = setmetatable({}, self);
	self.__index = self;
	
	o.name = name;	
	o.profile = profile or self.defaults();	
	o.effects = {};

	o:SetupType();
	o:CreateFrame();
	
	return o;
end

function AnchorProto:Release()
	for guid, guidEffects in pairs(self.effects) do
		for spellId, effect in pairs(guidEffects) do
			effect:Release();
		end
	end
	
	self:Hide();
end

function AnchorProto:SetupType()
	if(self.profile.type == "ungrouped") then
		self.layout_ApplyProfile = NoOp;
		self.layout_ReanchorFrames = ungrouped_ReanchorFrames;
		self.layout_SortEffects = NoOp;
		self.layout_NewEffect = ungrouped_NewEffect;
		self.layout_RemoveEffect = ungrouped_RemoveEffect;
		self.layout_Clear = NoOp;
		
		self.UnitUpdated = NoOp;
		
	elseif(self.profile.type == "byUnit") then
		self.layout_ApplyProfile = byUnit_ApplyProfile;
		self.layout_ReanchorFrames = byUnit_ReanchorEffects;
		self.layout_SortEffects = byUnit_SortEffects;
		self.layout_NewEffect = byUnit_NewEffect;
		self.layout_RemoveEffect = byUnit_RemoveEffect;
		self.layout_Clear = byUnit_Clear;
		
		self.UnitUpdated = byUnit_UnitUpdated;
		
		self.units = {};
	
	elseif(self.profile.type == "bySpell") then
	end
end

function AnchorProto:CreateFrame()
	local frame = CreateFrame("Frame", "SF_Anchor_"..(self.name), UIParent);
	self.frame = frame;
	
	-- This has to go, probably I will merge the frame object into
	-- the AnchorProto object itself
	frame.profile = self.profile;	
	
	frame:EnableMouse(true);
	frame:SetScript("OnDragStart", self.frame.StartMoving);
	frame:SetScript("OnDragStop", OnDragStop);	
	
	frame:SetBackdrop(backdrop);
	frame:SetBackdropColor(0, 0, 0, 0.5);
	frame:SetBackdropBorderColor(1, 1, 1, 0.6);
	
	frame.label = frame:CreateFontString(nil, "OVERLAY");
	frame.label:SetPoint("LEFT", frame, "LEFT", 6, 0);
	frame.label:SetFontObject(GameFontHighlightSmall);
		
	frame:Hide();
	
	self:ApplyProfile();
	self:ApplyFilter();
		
	self.lastChild = self;
end

function AnchorProto:ApplyProfile()
	local profile = self.profile;
	local frame = self.frame;	

	frame:SetScale(profile.scale);		
	frame:SetWidth(profile.width);
	frame:SetHeight(profile.anchorHeight);
	frame:SetAlpha(profile.anchorAlpha);
	
	frame.label:SetText(self.name);
	
	frame:ClearAllPoints();
	frame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", profile.posX, profile.posY);	
	
	for guid, guidEffects in pairs(self.effects) do
		for spellId, effect in pairs(guidEffects) do
			effect:ApplyProfile();
		end
	end
	
	self:layout_ApplyProfile();
	
	self:Hide();
	self:ShowIfVisible();
end

-- ONLY TO BE USED WITH BYGROUP
local function LinkageFailed(msg, event, after)
	if(after) then
		DEFAULT_CHAT_FRAME:AddMessage("after "..event..": "..msg);
	else
		DEFAULT_CHAT_FRAME:AddMessage("before "..event..": "..msg);
	end
end

-- ONLY TO BE USED WITH BYGROUP
function AnchorProto:CheckLinkage(...)
	-- We start out with the anchor's lastChild, check the effect linkage within
	-- and then go to the previous header until we're back at the anchor
	
	local i = 1;
	local j = 1;

	local curHeader = self.lastChild;
	local curEffect;
	
	while(curHeader ~= self) do
		curEffect = curHeader.next;
		
		if(not curEffect) then
			return LinkageFailed(curHeader.frame:GetName()..".next == nil", ...);
		end
		
		if(curEffect.prev ~= curHeader) then
			return LinkageFailed(curEffect.frame:GetName()..".prev ~= "..curHeader.frame:GetName(), ...);
		end
		
		j = 1;
		
		while(true) do
			if(curEffect.next) then				
				if(curEffect.next.prev ~= curEffect) then
					return LinkageFailed(curEffect.frame:GetName()..".next.prev ~= "..curEffect.frame:GetName(), ...);
				end
				
				curEffect = curEffect.next;
			else
				if(curHeader.lastChild ~= curEffect) then
					return LinkageFailed(curHeader.frame:GetName()..".lastChild ~= "..curEffect.frame:GetName(), ...);
				end
				
				if(curEffect.alignedHeader and not curHeader.nextHeader) then
					return LinkageFailed(curHeader.frame:GetName()..".nextHeader == nil, but last child "..curEffect.frame:GetName()..".alignedHeader == "..curEffect.alignedHeader.frame:GetName(), ...);
				end
				
				if(curHeader.nextHeader and not curEffect.alignedHeader) then
					return LinkageFailed(curHeader.frame:GetName()..".nextHeader ~= nil, but last child "..curEffect.frame:GetName()..".alignedHeader == nil", ...);
				end
				
				if(curEffect.alignedHeader ~= curHeader.nextHeader) then
					return LinkageFailed(curHeader.frame:GetName()..".nextHeader ~= "..curEffect.frame:GetName()..".alignedHeader", ...);
				end
		
				break;
			end
		
			j = j + 1;
			
			if(j == 100) then
				return LinkageFailed("Reached j == 100 on "..curHeader.frame:GetName(), ...);
			end
		end
		
		if(curHeader.prevHeader.nextHeader ~= curHeader and curHeader.prevHeader ~= self) then
			return LinkageFailed(curHeader.frame:GetName()..".prevHeader.nextHeader ~= "..curHeader.frame:GetName(), ...);
		end
		
		curHeader = curHeader.prevHeader;
		
		i = i + 1;
		
		if(i == 100) then
			return LinkageFailed("Reached i == 100", ...);
		end
	end
end

function AnchorProto:ApplyFilter()
	local profile = self.profile;

	if(profile.unitFilterMode == "inclusive") then
		self.unitRelativeFilter_Check = unitRelativeFilter_CheckInclusive;
	else
		self.unitRelativeFilter_Check = unitRelativeFilter_CheckExclusive;
	end
end

function AnchorProto:Lock()
	self.frame:SetMovable(false);
	self.frame:RegisterForDrag();
	
	self:Hide();
	self:ShowIfVisible();
end

function AnchorProto:Unlock()
	self.frame:SetMovable(true);	
	self.frame:RegisterForDrag("LeftButton");	
	
	self:Show(true);
end

function AnchorProto:EffectApplied(sourceGUID, sourceName, destGUID, destName, spellId, amount, quality, duration)
	local profile = self.profile;
	
	if((sourceGUID ~= playerGUID) and (profile.filterAppliedByMe)) then return; end
	if((destGUID ~= playerGUID) and (profile.filterAppliedOnMe)) then return; end
	
	if(not self:unitRelativeFilter_Check(destGUID)) then return; end

	local effect = ShieldFrames.EffectProto:New(self);
	
	if(not self.effects[destGUID]) then
		self.effects[destGUID] = {[spellId] = effect};
	else
		self.effects[destGUID][spellId] = effect;
	end	
	
	effect.sourceName = sourceName;
	effect.destGUID = destGUID;
	effect.destName = destName;
	effect.spellId = spellId;
	
	self:layout_NewEffect(effect, destGUID, destName, spellId);
	
	effect:SetupValues(amount, quality, duration);
	
	effect.frame:Show();
end

function AnchorProto:EffectUpdated(guid, spellId, amount, quality, duration)
	local guidEffects = self.effects[guid];

	if(guidEffects and guidEffects[spellId]) then
		guidEffects[spellId]:UpdateValues(amount, quality, duration);
	end
end

function AnchorProto:EffectRemoved(guid, spellId)
	local guidEffects = self.effects[guid];
	
	if(guidEffects and guidEffects[spellId]) then
		local effect = guidEffects[spellId];
		
		self:layout_RemoveEffect(effect, guid, spellId);
		
		effect:Release();
		
		guidEffects[spellId] = nil;
	end
end

function AnchorProto:UnitUpdated(guid, amount, quality)
end

function AnchorProto:UnitCleared(guid)
	local guidEffects = self.effects[guid];
	
	if(guidEffects) then
		for spellId, effect in pairs(guidEffects) do
			-- A little inefficient here, should optimize that later
			-- But then again, it's rare to have such a message on a
			-- non-empty unit
			self:Absorbs_EffectRemoved(guid, spellId);
		end
	end
end

function AnchorProto:AreaCreated(sourceGUID, sourceName, triggerGUID, spellId, value, quality)
	self:EffectApplied(sourceGUID, sourceName, triggerGUID, "Area Effect by "..sourceName, spellId, value, quality, nil);

	AreaTriggerSpells[triggerGUID] = spellId;
end

function AnchorProto:AreaUpdated(triggerGUID, value)
	self:EffectUpdated(triggerGUID, AreaTriggerSpells[triggerGUID], amount, quality, duration);
end

function AnchorProto:AreaCleared(triggerGUID)
	self:EffectRemoved(triggerGUID, AreaTriggerSpells[triggerGUID]);

	AreaTriggerSpells[triggerGUID] = nil;
end

function AnchorProto:Clear()
	for guid, guidEffects in pairs(self.effects) do
		for spellId, effect in pairs(guidEffects) do
			effect.frame:ClearAllPoints();
			
			effect:Release();
		end
	end
	
	self:layout_Clear();
	
	self.effects = {};
end

function AnchorProto:ShowIfVisible()
	self:Show();
end

function AnchorProto:Show(force)
	if(self.profile.anchorVisible or force) then
		self.frame:Show();
	end
end

function AnchorProto:Hide()
	self.frame:Hide();
end
