
------------------------------------------------------------------------
--
-- ShieldFrames
--
-- Copyright (C) 2010  Philipp Schmidt
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to:
--
-- Free Software Foundation, Inc.,
-- 51 Franklin Street, Fifth Floor,
-- Boston, MA  02110-1301, USA.
--
--
------------------------------------------------------------------------


local ShieldFrames = _G.ShieldFrames;
if(not ShieldFrames) then return; end

local EffectProto = {};
ShieldFrames.EffectProto = EffectProto;

local EffectFrameCounter = 1;
local EffectCache = {};


local function OnUpdateDuration(self)
	local t = GetTime();
	
	if(t > self.timeout) then
		self:SetValue(0);
		self:SetScript("OnUpdate", nil);
	else
		self:SetValue(self.timeout - t);
	end
end

function EffectProto:New(anchor)
	if(#EffectCache > 0) then
		local cachedEffect = tremove(EffectCache, 1);
		
		cachedEffect:ApplyProfile();
		
		return cachedEffect;
	end

	local o = setmetatable({}, self);
	self.__index = self;
	
	o.anchor = anchor;
	o.profile = anchor.profile;
	
	o:CreateFrame();
	
	return o;
end

function EffectProto:Release()
	self.frame:Hide();
	
	self.frame:ClearAllPoints();
	
	self.leftText:SetText("");
	self.centerText:SetText("");
	self.rightText:SetText("");

	self.frame:SetScript("OnUpdate", nil);	
	self.amount = nil;
	
	tinsert(EffectCache, self);
end

function EffectProto:CreateFrame()
	local frame = CreateFrame("Frame", "SF_Effect"..EffectFrameCounter, UIParent);
	self.frame = frame;
	
	EffectFrameCounter = EffectFrameCounter + 1;
	
	frame:Hide();
			
	local icon = frame:CreateTexture(nil, "ARTWORK");
	self.icon = icon;
	
	icon:SetPoint("TOPLEFT", frame);
	icon:SetPoint("BOTTOMLEFT", frame);
	icon:SetTexCoord(0.07, 0.93, 0.07, 0.93);
	
	local amountBar = CreateFrame("StatusBar", nil, frame);
	self.amountBar = amountBar;
	
	amountBar:SetStatusBarTexture([[Interface\Addons\ShieldFrames\BantoBar]]);
	amountBar:SetPoint("TOPRIGHT", frame);
	amountBar:SetOrientation("HORIZONTAL");
	
	local amountBg = amountBar:CreateTexture(nil, "BACKGROUND");
	self.amountBg = amountBg;
	
	amountBg:SetTexture([[Interface\Addons\ShieldFrames\BantoBar]]);
	amountBg:SetAllPoints();
	
	local leftText = amountBar:CreateFontString(nil, "OVERLAY", "GameFontHighlightSmallOutline");
	self.leftText = leftText;
	
	leftText:SetFontObject(GameFontHighlightSmallOutline);
	leftText:SetPoint("LEFT", amountBar, "LEFT", 2, 0);
	
	local centerText = amountBar:CreateFontString(nil, "OVERLAY");
	self.centerText = centerText;
	
	centerText:SetFontObject(GameFontHighlightSmallOutline);
	centerText:SetPoint("CENTER", amountBar, "CENTER");
	
	local rightText = amountBar:CreateFontString(nil, "OVERLAY");
	self.rightText = rightText;

	rightText:SetFontObject(GameFontHighlightSmallOutline);
	rightText:SetPoint("RIGHT", amountBar, "RIGHT", -2, 0);
	
	local durationBar = CreateFrame("StatusBar", nil, frame);
	self.durationBar = durationBar;
	
	durationBar:SetStatusBarTexture([[Interface\Addons\ShieldFrames\Flat]]);
	durationBar:SetStatusBarColor(0, 0, 1, 1.0);
	
	durationBar:SetPoint("BOTTOMRIGHT", frame);
	durationBar:SetOrientation("HORIZONTAL");
	
	local durationBg = durationBar:CreateTexture(nil, "BACKGROUND");
	self.durationBg = durationBg;
	
	durationBg:SetTexture([[Interface\Addons\ShieldFrames\Flat]]);
	durationBg:SetVertexColor(0.6, 0.6, 0.6, 0.8);
	durationBg:SetAllPoints();
	
	self:ApplyProfile();
end

function EffectProto:ApplyProfile()
	local profile = self.profile;

	self.frame:SetScale(profile.scale);		
	self.frame:SetWidth(profile.width);
	self.frame:SetHeight(profile.effectHeight);
	self.frame:SetAlpha(profile.effectAlpha);
	
	self.icon:SetWidth(profile.effectHeight);
	self.icon:SetHeight(profile.effectHeight);
	
	if(profile.iconVisible) then
		self.amountBar:SetPoint("TOPLEFT", self.icon, "TOPRIGHT");
		self.durationBar:SetPoint("BOTTOMLEFT", self.icon, "BOTTOMRIGHT");
	
		self.icon:Show();
	else
		self.amountBar:SetPoint("TOPLEFT", self.frame, "TOPLEFT");
		self.durationBar:SetPoint("BOTTOMLEFT", self.frame, "BOTTOMLEFT");
	
		self.icon:Hide();
	end
	
	if(profile.durationVisible) then
		self.amountBar:SetPoint("BOTTOMLEFT", self.durationBar, "TOPLEFT", 0, 1);
		self.amountBar:SetPoint("BOTTOMRIGHT", self.durationBar, "TOPRIGHT", 0, 1);	
	
		self.durationBar:Show();		
	else
		self.amountBar:SetPoint("BOTTOMLEFT", self.frame, "BOTTOMLEFT");
		self.amountBar:SetPoint("BOTTOMRIGHT", self.frame, "BOTTOMLEFT");
	
		self.durationBar:Hide();
	end
	
	local path, size, style = GameFontHighlightSmallOutline:GetFont();
	size = size + profile.effectTextSize;
	
	self.leftText:SetFont(path, size, style);
	self.centerText:SetFont(path, size, style);
	self.rightText:SetFont(path, size, style);		
	
	self.durationBar:SetHeight(profile.durationHeight);

	if(self.amount) then
		self:UpdateStaticText();
		self:UpdateDynamicText();
	end
end

function EffectProto:SetupValues(amount, quality, duration)
	local spellName, _, spellIcon = GetSpellInfo(self.spellId);	
	
	self.spellName = spellName;
	self.icon:SetTexture(spellIcon);
	
	-- Force static text update	and sorting
	-- Infinite effects start always at -1, we get the same
	-- effect on thise
	self.max = -10;
	self:UpdateValues(amount, quality, duration);	
	
	if(amount < 1) then
		-- We have to "correct" the errors UpdateValues made in case of amount == 0
		-- and set the bar in case amount < 1 (infinite amounts)
		-- The difference between these two cases is that we want dynamic text on the first
		-- one, but not the latter
		-- We need at least one update of the dynamic text as well
		self.amountBar:SetMinMaxValues(0, 1);
		self.amountBar:SetValue(1);
		
		self:UpdateDynamicText();
	end
end

function EffectProto:UpdateStaticText()
	local max = (self.amount > 0) and self.max or "-";

	-- left text	
	local str = self.profile.leftFormat or "";	
	self.leftDynamic = (strfind(str, "%%[ap]")) and true or false;
	
	str = gsub(str, "%%s", self.sourceName);
	str = gsub(str, "%%d", self.destName);
	str = gsub(str, "%%n", self.spellName);
	str = gsub(str, "%%m", max);	
	
	self.leftTpl = str;
	self.leftText:SetText(str);
	
	-- center text	
	str = self.profile.centerFormat or "";
	self.centerDynamic = (strfind(str, "%%[ap]")) and true or false;
	
	str = gsub(str, "%%s", self.sourceName);
	str = gsub(str, "%%d", self.destName);
	str = gsub(str, "%%n", self.spellName);
	str = gsub(str, "%%m", max);		
	
	self.centerTpl = str;	
	self.centerText:SetText(str);
	
	-- right text	
	str = self.profile.rightFormat or "";
	self.rightDynamic = (strfind(str, "%%[ap]")) and true or false;
	
	str = gsub(str, "%%s", self.sourceName);
	str = gsub(str, "%%d", self.destName);
	str = gsub(str, "%%n", self.spellName);
	str = gsub(str, "%%m", max);		
	
	self.rightTpl = str;
	self.rightText:SetText(str);
end

function EffectProto:UpdateDynamicText()
	local str;
	
	local amount, percent;
	
	if(self.max > 0) then
		amount = self.amount;
		percent = format("%.0f", (amount / self.max) * 100);
		
	elseif(self.max == 0) then	
		amount = 0;
		percent = 100;
	else
		amount = "-";
		percent = "-";
	end
	
	if(self.leftDynamic) then
		str = self.leftTpl;	
		str = gsub(str, "%%a", amount);
		str = gsub(str, "%%p", percent);
		
		self.leftText:SetText(str);
	end
	
	if(self.centerDynamic) then
		str = self.centerTpl;
		str = gsub(str, "%%a", amount);
		str = gsub(str, "%%p", percent);
		
		self.centerText:SetText(str);
	end
	
	if(self.rightDynamic) then
		str = self.rightTpl;
		str = gsub(str, "%%a", amount);
		str = gsub(str, "%%p", percent);
		
		self.rightText:SetText(str);
	end
end

function EffectProto:UpdateValues(amount, quality, duration)
	self.amount = amount;
	
	if(quality) then
		self.amountBar:SetStatusBarColor(1 - quality, quality, 0, 1.0);		
		self.amountBg:SetVertexColor(0.8 - quality * 0.2, 0.5 + quality * 0.1, 0.6, 0.6);
	end		
	
	if(duration) then
		self.durationBar:SetScript("OnUpdate", nil);
		self.durationBar:SetMinMaxValues(0, duration);
		self.durationBar:SetValue(duration);	
		self.durationBar.duration = duration;
		self.durationBar.timeout = GetTime() + duration;
		self.durationBar:SetScript("OnUpdate", OnUpdateDuration);
	end	
	
	if(amount > self.max) then
		self.max = amount;
		self.amountBar:SetMinMaxValues(0, amount);
	
		self:UpdateStaticText();
		
		self.anchor:layout_SortEffects(self.destGUID, self.spellId);
	end		
	
	if(amount > 0) then
		self.amountBar:SetValue(amount);	
		self:UpdateDynamicText();		
	end
end

-- alignObject requires a frame property to align to, it can be an Anchor, Header or Effect
function EffectProto:AlignTo(alignObject, reverseGrowth)
	local profile = self.profile;
	
	self.frame:ClearAllPoints();

	if(reverseGrowth) then
		self.frame:SetPoint("BOTTOMLEFT", alignObject.frame, "TOPLEFT", 0, profile.effectSpacing);		
	else
		self.frame:SetPoint("TOPLEFT", alignObject.frame, "BOTTOMLEFT", 0, -1 * profile.effectSpacing);
	end
end
